/**
 * @file DeepPass.cc
 * Implementation of the DeepPass skill.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.8 $
 * $Author: aip13 $
 * $Date: 2001/10/30 18:46:06 $
 */

#include "DeepPass.h"
#include "KickCommand.h"
#include "KickBallToAbsoluteSpeedVector.h"
#include "ParameterView.h"
#include "AgentView.h"
#include "BallView.h"
#include "PlayerView.h"
#include "GameView.h"
#include "TVector.h"
#include "MarkerView.h"


RS_BEGIN_NAMESPACE


// --- CONSTRUCTORS ---

/**
 * Constructor.
 */

DeepPass::DeepPass()
  : SkillWithArguments(),
    parameterView(NULL), 
    markerView(NULL),
    agentView(NULL),
    ballView(NULL),
    playerView(NULL),
    gameView(NULL)
{
  getView(parameterView, PARAMETER_VIEW_ID);
  getView(markerView, MARKER_VIEW_ID, parameterView);
  getView(agentView, AGENT_VIEW_ID, parameterView);
  getView(ballView, BALL_VIEW_ID);
  getView(playerView, PLAYER_VIEW_ID);
  getView(gameView, GAME_VIEW_ID);

  
  /* Default values
     A good value for the length of the vector is between 10 and 20.
     The angle of the vector should be related to the direction of
     the recieving player. */
  setVector(Vector(AngleDeg(0.0), 0.0));
  setSpeed(0.3);

  /* We use lastCalculationTime to avoid calculating the same thing
     twice in the same cycle */
  lastCalculationTime = 0;
  

#ifdef DEEPPASS_DEBUG
  fp = fopen("/home/aip13/logfile/DeepPass.log", "w");
#endif
}

// --- DESTRUCTOR ---

DeepPass::~DeepPass()
{
  viewManager->releaseView(PARAMETER_VIEW_ID);
  viewManager->releaseView(MARKER_VIEW_ID);
  viewManager->releaseView(AGENT_VIEW_ID);
  viewManager->releaseView(BALL_VIEW_ID);
  viewManager->releaseView(PLAYER_VIEW_ID);
  viewManager->releaseView(GAME_VIEW_ID);

#ifdef DEEPPASS_DEBUG
  fclose(fp);
#endif
}


// --- METHODS ---
/**
 * Sets the arguments to the skill.
 * The skill wants two arguments.
 * 1. An absolute Vector from the player closest to goal, defining 
 *    to where we should pass the ball.
 * 2. The speed the ball should have in the point, defined by
 *    the Vector above.
 * The argument elments are:
 * arguments[0] = The angle of the Vector
 * arguments[1] = The length of the Vector
 * arguments[2] = The speed of the ball.
 * 
 * A good value for the length of the vector is between 10 and 20.
 * The angle of the vector should be related to the direction of
 * the recieving player. 
 * 
 * If no arguments are passed, default values will be used.
 */
void DeepPass::setArguments(const std::vector<Float>& arguments)
  throw(RS::TooFewArgumentsException)
{
  setVector(Vector(AngleDeg(arguments[0]), arguments[1]));
  setSpeed(arguments[2]);
}

/** Used to set the skill's speed argument to @a spd. */
void DeepPass::setSpeed(const Float spd)
{
  speed = spd;
}

/** Used to get the skill's speed argument. */
Float DeepPass::getSpeed() const
{
  return speed;
} 

/** Used to set the skill's vector argument to @a vec. */
void DeepPass::setVector(const Vector& vec)
{
  vector = vec;
}

/** Used to get the skill's vector argument. */
Vector DeepPass::getVector() const
{
  return vector;
}

/** Used to get the player that we intend to pass to */
PlayerObject DeepPass::getPlayerToPassTo() const
{
  return playerToPassTo;
}

/** Used to get the calculated pass vector */
Vector DeepPass::getPassVector()
{
  return passVector;
}


/**
 * Same as
 * setArguments(foo);
 * generatePlan();
 */

Plan DeepPass::generatePlan(const std::vector<Float>& arguments) {
  setArguments(arguments);
  return generatePlan();
}

/**
 * Calculates the pass vector according to the skill's arguments and
 * calculates how hard we should kick the ball.
 */
Vector DeepPass::calculatePassVector()
{
  playerToPassTo = getTeammateClosestToGoal();
  
  /* Calculates the vector between the Agent and the target */
  Vector passvector(0,0);
  if (playerToPassTo.getVector().isKnown()) {
    passvector = agentView->getAgentObject(NOW).getAbsoluteVectorTo(playerToPassTo.getPosition()).getVector();
    /* Add the Vector and the speed of the ball supplied to the skill. And
       calculate how hard we should kick. */
    passvector = recalculatePassVector(passvector);
  } else {
#ifdef DEEPPASS_DEBUG
    fprintf(fp,"Error:we can see the player but we don't know where he is...\n");
#endif
  }

  return passvector;
}

/**
 * Generates a plan for achieving a pass.
 * Chooses the teammate closest to goal that we actually can 
 * pass to.
 */

Plan DeepPass::generatePlan()
{
  if (lastCalculationTime != gameView->getCurrentTime().getT()) {
    /* If we didn't already calculate the passVector this cycle,
       calculate it. Though it is done in applicable(), so it should
       already be calculated.... */
    tm = playerView->getTeammates(NOW);
    
    findPlayerToPassToAndCalculatePassVector();
  }

#ifdef DEEPPASS_DEBUG
  fprintf(fp, "Absolute vector to player: angle: %f, speed: %f\n", passVector.getAngle().getDeg(), passVector.getLength());
#endif

  KickBallToAbsoluteSpeedVector kickBallToAbsoluteSpeedVector(passVector);
  return kickBallToAbsoluteSpeedVector.generatePlan();
}


/**
 * Calculates which player to pass to. We find a player that we
 * can pass to and that is closest to the goal. 
 * We return a fuzzy bool indicating if we found a player to pass
 * to or not.
 */
FuzzyBool DeepPass::findPlayerToPassToAndCalculatePassVector()
{
  //tm = playerView->getTeammates(NOW);
  KickBallToAbsoluteSpeedVector kickBallToAbsoluteSpeedVector(passVector);
  FuzzyBool canPass(0.0);
  //int i = 10;
  while (tm.size() != 0 && !canPass) {// i > 0) {
    //i--;
  /* Calculate a new pass vector to the player closest to goal. 
     In calculatePassVector, the player that we want to pass to
     is removed from tm, so that we don't find him again. Otherwise
     we would loop forever.
  */
    passVector = calculatePassVector();
    kickBallToAbsoluteSpeedVector.setVector(passVector);
 
    /* Check if it is possible to pass where we want */
    canPass = kickBallToAbsoluteSpeedVector.applicable() && 
      FuzzyBool(isPassForward(passVector));
  }
  return canPass;
}

/**
 * Returns the teammate closest to goal. This is the player intended to
 * recieve the pass.
 */
PlayerObject DeepPass::getTeammateClosestToGoal()
{
  /* If we don't see any teammates, the skill isn't applicable, therefore
     we can initialize teamMate to the first teammate we see. */
  PlayerObject teamMate = tm[0];
  std::vector<PlayerObject>::iterator iter = tm.begin();
  std::vector<PlayerObject>::iterator closestPlayerIter = iter;
  for (unsigned int i=0;i<tm.size();) {
    PlayerObject aPlayer = tm[i];
    if (distanceToGoal(aPlayer) < distanceToGoal(teamMate)) {
      teamMate = aPlayer;
      
      /* Delete the player closest to goal from the teammate vector, since
	 we may have to find a new player closest to goal if we can't pass
	 to this one */
      /* Though we don't delete him yet, we only save the position */
      closestPlayerIter = iter;
    } else {
      i++;
      iter++;
    }
  }
  tm.erase(closestPlayerIter);
  return teamMate;
}

/**
 * Calculates the distance between a player and the oppononent's goal
 */
Float DeepPass::distanceToGoal(PlayerObject &aPlayer)
{
  PointUCO goalPosition = markerView->getTheirGoal(NOW).getPosition();
  
if (aPlayer.getPosition().isKnown() && goalPosition.isKnown())
    return goalPosition.getDistanceTo(aPlayer.getPosition()).getValue();
  else
    return 100.0;/* If we don't know where the player is, return a big value */

}

/**
 * Used to calculate how we should pass according to the arguments
 * supplied to the skill.
 */
Vector DeepPass::recalculatePassVector(Vector& passvec)
{
  TVector tpassvec(passvec);
  TVector tvector(vector);
  /* Add the argument vector to the pass vector */
  tpassvec += tvector;
  passvec.setLength(tpassvec.getLength());
  passvec.setAngle(tpassvec.getAngle());

  Vector tmpvec = passvec;
  tmpvec.setLength(1);
  /* Calculate how hard we should kick */
  passvec = tmpvec * (speed + passvec.getLength() * (1 - parameterView->SP_ball_decay));
  return passvec;
}



/**
 * Persistent.
 * FALSE
 */
FuzzyBool DeepPass::persistent()
{
  /* Never persistent. */
  return FuzzyBool(0.0);
}

/**
 * Same as:
 * setArguments(foo);
 * applicable();
 */
FuzzyBool DeepPass::applicable(const std::vector<Float>& arguments) {
  setArguments(arguments);
  return applicable();
}


/**
 * Checks if the ball is in front of the agent
 */
FuzzyBool DeepPass::isBallInFrontOfAgent()
{
  PointUCO ballPosition;

  if (ballView->getBallPosition(NOW).isKnown())
    ballPosition = ballView->getBallPosition(NOW);
  else
    return FuzzyBool(0.0);
  
  VectorUCO ballVector = agentView->getAgentObject(NOW).getRelativeVectorTo(ballPosition);
  
  /* Here we check if the ball is in front of us.
     Maybe we should check for a more narrow angle */
  return (ballVector.getAngle() > -90) && (ballVector.getAngle() < 90);
}

/**
 * Checks if the agent's body is turned towards the opponent's side
 */
FuzzyBool DeepPass::isAgentHeadingToOpponentGoal()
{
  return (agentView->getBodyDirection(NOW) > -90) && 
         (agentView->getBodyDirection(NOW) < 90);
}

/**
 * Checks if the ball is within kickable range
 */
FuzzyBool DeepPass::isBallKickable()
{
  PointUCO ballPosition;
  PointUCO agentPosition;
  
  if (ballView->getBallPosition(NOW).isKnown()) {
    ballPosition = ballView->getBallPosition(NOW);
  } else {
    return FuzzyBool(0.0);
  }
  agentPosition = agentView->getAgentPosition(NOW);
  
  if(agentView->getAgentPosition(NOW).isKnown()) {

#ifdef DEEPPASS_DEBUG
    fprintf(fp, "isBallKickable: Agentposition: %f,%f\n", agentPosition.getX().getValue(), agentPosition.getY().getValue());
#endif
    
    return agentView->getAgentPosition(NOW).getDistanceTo(ballPosition).getValue() < parameterView->SP_kickable_area;
  } else {
    return FuzzyBool(0.0);
  }
}

/** Checks if we pass forward */
FuzzyBool DeepPass::isPassForward(Vector& passvec)
{
  return (passvec.getAngle() > -90) && (passvec.getAngle() < 90);
}  


/**
 * Returns the applicable value
 */
FuzzyBool DeepPass::applicable()
{
  static FuzzyBool applicable;
  if (lastCalculationTime == gameView->getCurrentTime().getT()) {
    return applicable;
  }
  else {
    lastCalculationTime = gameView->getCurrentTime().getT();
    tm = playerView->getTeammates(NOW);
  
    /* Can we see any teammates? */
    if (tm.size() == 0) {
#ifdef DEEPPASS_DEBUG
      fprintf(fp, "tm.size() == 0\n");
#endif
      applicable = FuzzyBool(0.0);
      /* Do we know where we are? */
    } else if ((agentView->getAgentPosition(NOW)).isUnknown()) {
#ifdef DEEPPASS_DEBUG
      fprintf(fp, "Unknown agent position\n");
#endif
      applicable = FuzzyBool(0.0);
      /* Is the ball in front of us? */
    } else if (!isBallInFrontOfAgent()) {
#ifdef DEEPPASS_DEBUG
      fprintf(fp, "Ball is not in front of me\n");
#endif
      applicable = FuzzyBool(0.0);
      /* Are we facing our opponent's goal? */
    } else if (!isAgentHeadingToOpponentGoal()) {
#ifdef DEEPPASS_DEBUG
      fprintf(fp, "I'm not heading to the opponent's goal\n");
#endif
      applicable = FuzzyBool(0.0);
      /* Can we kick the ball? */
    } else if (!isBallKickable()) { 
#ifdef DEEPPASS_DEBUG
    fprintf(fp, "I can't kick the ball\n");
#endif
    applicable = FuzzyBool(0.0);
    } else {
      FuzzyBool canPass = findPlayerToPassToAndCalculatePassVector();
      if (canPass) {
#ifdef DEEPPASS_DEBUG
	fprintf(fp, "I can do it, go long\n");
#endif
	//FIXME: calculate how sure we are to complete the pass
	applicable = FuzzyBool(1.0);
      } else {
	applicable = FuzzyBool(0.0);
      }
    }
  }
  return applicable;
}

RS_END_NAMESPACE
