/* -*- C++ -*- */
   
#ifndef DEEPPASS_H
#define DEEPPASS_H


#include "Skill.h"
#include "SkillWithArguments.h"
#include "PlayerObject.h"
#include "CoordVector.h"
#include "CoordPoint.h"
#include "view_constants.h"
#include "rs_log.h"

#ifdef DEEPPASS_DEBUG
#include <stdio.h>
#endif


RS_BEGIN_NAMESPACE


class ParameterView;
class AgentView;
class BallView;
class PlayerView;
class GameView;
class MarkerView;


/**
 * @class DeepPass DeepPass.h
 * This skill is used to pass the ball forward as far as possible.
 * It starts by checking if we are able to pass to the player closest
 * to the opponent's goal. If not we check with the second closest
 * player and so on, until we either find a player to pass to, or
 * decide that the skill isn't applicable.
 * 
 * The skill can take two arguments.
 * 1. An absolute Vector from the player closest to goal, defining 
 *    to where we should pass the ball.
 * 2. The speed the ball should have in the point, defined by
 *    the Vector above.
 * The argument elments are:
 * arguments[0] = The angle of the Vector
 * arguments[1] = The length of the Vector
 * arguments[2] = The speed of the ball.
 * 
 * A good value for the length of the vector is between 10 and 20.
 * The angle of the vector should be related to the direction of
 * the recieving player. 
 * 
 * If no arguments are passed, default values will be used.
 *
 *
 * @pre 
 *  - See at least one teammate. 
 *  - Body turned to opponent's goal. 
 *  - Can kick ball. Ball in front of us. 
 *  - Be able to set the ball to the desired speed in the calculated point, to which we want to pass to.
 * @post Ball hopefully at the player closest to goal.
 *
 * @todo Applicable should return a value between 0 and 1, indicating how sure we are that our player actually gets the ball.
 * 
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.5 $
 * $Author: aip13 $
 * $Date: 2001/10/30 18:46:06 $
 */

class DeepPass : public SkillWithArguments {
public:
  /** 
   * Default constructor. Here you can initialize the skill. 
   */
  DeepPass();

  /** 
   * Destructor should clean up after the skill. 
   */
  virtual ~DeepPass();

  
  /** 
   * Generates a plan for achieving a pass.
   * Chooses the teammate closest to goal that we actually can 
   * pass to.
   */
  virtual Plan generatePlan();


  /** 
   * This method should return a FuzzyBool telling the caller
   * if the skill wants to be called again. 
   */
  virtual FuzzyBool persistent();

  /** 
   * This method should implement the precondition specified in
   * the description of the skill. 
   */
  virtual FuzzyBool applicable();
  
  /** Return @c true (close to 1) if the skill has succeeded,
      i.e. its post conditions are satisfied. */
  virtual FuzzyBool succeed() { return FuzzyBool(0.5); };

  /** Return @c true (close to 1) if the skill has failed,
      i.e. something went wrong. */
  virtual FuzzyBool failed() { return FuzzyBool(0.5); };

  /** Used to set the skill's speed argument to @a spd. */
  virtual void setSpeed(const Float spd);
  /** Used to get the skill's speed argument. */
  virtual Float getSpeed() const;

  /** Used to set the skill's vector argument to @a vec. */
  virtual void setVector(const Vector& vec);
  /** Used to get the skill's vector argument. */
  virtual Vector getVector() const;

  /** Used to get the player that we intend to pass to */
  PlayerObject getPlayerToPassTo() const;
  
  /** Used to get the calculated pass vector */
  Vector getPassVector();
 
  /**
   * Calculates the pass vector according to the skill's arguments and
   * calculates how hard we should kick the ball.
   */
  Vector calculatePassVector();

  /**
   * Calculates which player to pass to. We find a player that we
   * can pass to and that is closest to the goal. 
   * We return a fuzzy bool indicating if we found a player to pass
   * to or not.
   */
  FuzzyBool findPlayerToPassToAndCalculatePassVector();

  /** Used to set the skill's arguments in a generic way.
   * This method must be overridden by subclasses.
   * It should transform the generic vector of floats to
   * the actual arguments.
   * It will throw TooFewArgumentsException if not enough arguments
   * are given. 
   * The skill wants two arguments.
   * 1. An absolute Vector from the player closest to goal, defining 
   *    to where we should pass the ball.
   * 2. The speed the ball should have in the point, defined by
   *    the Vector above.
   * The argument elments are:
   * arguments[0] = The angle of the Vector
   * arguments[1] = The length of the Vector
   * arguments[2] = The speed of the ball.
   * 
   * A good value for the length of the vector is between 10 and 20.
   * The angle of the vector should be related to the direction of
   * the recieving player. 
   * 
   * If no arguments are passed, default values will be used.
   */
  virtual void setArguments(const std::vector<Float>& arguments)
    throw(RS::TooFewArgumentsException);


  /** This method is used to first set the arguments of the skill
    and then generate a plan for how the goal of the skill should
    be obtained by the agent. It uses the methods setArguments and
    generatePlan(), so it should not be overriden unless it should
    made unavailable. */
  virtual Plan generatePlan(const std::vector<Float>& arguments);
  virtual FuzzyBool applicable(const std::vector<Float>& arguments);

protected:

  Float speed;
  Vector vector;
  Vector passVector;
  std::vector<PlayerObject> tm;
  PlayerObject playerToPassTo;
  ParameterView* parameterView;
  MarkerView* markerView;
  AgentView* agentView;
  BallView* ballView;
  PlayerView* playerView;
  GameView* gameView;
  UInt16 lastCalculationTime;

  /**
   * Checks if the ball is in front of the agent
   */
  FuzzyBool isBallInFrontOfAgent();
  
  /**
   * Checks if the agent's body is turned towards the opponent's side
   */
  FuzzyBool isAgentHeadingToOpponentGoal();


  /**
   * Checks if the ball is within kickable range
   */
  FuzzyBool isBallKickable();
 
  /** 
   * Checks if we pass forward 
   */
  FuzzyBool isPassForward(Vector& passvec);

  /**
   * Returns the teammate closest to goal. This is the player intended to
   * recieve the pass.
   */
  PlayerObject getTeammateClosestToGoal();

  /**
   * Calculates the distance between a player and the oppononent's goal
   */
  Float distanceToGoal(PlayerObject &aPlayer);

  /**
   * Used to calculate how we should pass according to the arguments
   * supplied to the skill.
   */
  Vector recalculatePassVector(Vector& passvec);



#ifdef DEEPPASS_DEBUG
  FILE *fp;
#endif

  
  /** Hide copy-constructor. */
  DeepPass(const DeepPass&);
  /** Hide assignment operator. */
  DeepPass& operator= (const DeepPass&);
};


RS_END_NAMESPACE

#endif // !DEEPPASS_H
