/* -*- Mode: C++ -*-
 * $Id: MakePassable.cc,v 1.2 2001/10/11 09:15:34 aip3 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: MakePassable.cc
 * Unitname: Framework/skill
 * $Revision: 1.2 $
 * Created by: Daniel Garpe, Karl Heijdenberg, Anders Lindgren 2001-10-08
 * Last modified by $Author: aip3 $ $Date: 2001/10/11 09:15:34 $
 *
 *
 * HISTORY
 */

/**
 * @file MakePassable.cc
 * Implementation file for the skill class MakePassable.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.2 $
 * $Author: aip3 $
 * $Date: 2001/10/11 09:15:34 $
 */


#include <math.h>
#include <AngleDegrees.h> 
#include "MakePassable.h"
#include "view_constants.h"
#include "TurnCommand.h"
#include "DashCommand.h"
#include "ParameterView.h"
#include "MarkerView.h"
#include "AgentView.h"
#include "BallView.h"
#include "GameView.h"
#include "PlayerView.h"
#include "rs_log.h"
#include "MoveToPositionFacingDirectionAvoidingObstacles.h"


RS_BEGIN_NAMESPACE


MakePassable::MakePassable()
  : Skill(),
    parameterView(NULL),
    markerView(NULL),
    agentView(NULL),
    ballView(NULL),
    gameView(NULL),
    playerView(NULL)
{
  parameterView = static_cast<ParameterView*>(viewManager->getView(PARAMETER_VIEW_ID));
  if ( parameterView == NULL ) {
    parameterView = new ParameterView(PARAMETER_VIEW_ID);
    viewManager->addView(parameterView);
  }

  markerView = static_cast<MarkerView*>(viewManager->getView(MARKER_VIEW_ID));
  if ( markerView == NULL ) {
    markerView = new MarkerView(MARKER_VIEW_ID, parameterView);
    viewManager->addView(markerView, true);
  }

  agentView = static_cast<AgentView*>(viewManager->getView(AGENT_VIEW_ID));
  if ( agentView == NULL ) {
    agentView = new AgentView(AGENT_VIEW_ID, parameterView);
    viewManager->addView(agentView, true);
  }

  ballView = static_cast<BallView*>(viewManager->getView(BALL_VIEW_ID));
  if ( ballView == NULL ) {
    ballView = new BallView(BALL_VIEW_ID);
    viewManager->addView(ballView, true);
  }
  
  gameView = static_cast<GameView*>(viewManager->getView(GAME_VIEW_ID));
  if ( gameView == NULL ) {
    gameView = new GameView(GAME_VIEW_ID);
    viewManager->addView(gameView, true);
  }
 
  playerView = static_cast<PlayerView*>(viewManager->getView(PLAYER_VIEW_ID));
  if ( playerView == NULL ) {
    playerView = new PlayerView(PLAYER_VIEW_ID);
    viewManager->addView(playerView, true);
  }
} 


MakePassable::~MakePassable()
{
  viewManager->releaseView(PARAMETER_VIEW_ID);
  viewManager->releaseView(MARKER_VIEW_ID);
  viewManager->releaseView(AGENT_VIEW_ID);
  viewManager->releaseView(BALL_VIEW_ID);
  viewManager->releaseView(GAME_VIEW_ID);
  viewManager->releaseView(PLAYER_VIEW_ID);
}


Plan MakePassable::generatePlan()
{
  static Plan plan;
  std::vector<PlayerObject> opponentList;
  std::vector<PlayerObject> teammateList;
  typedef std::vector<PlayerObject> Container;
  Container::iterator i;
  Container::iterator j;
  Container::iterator k;
  Container::iterator l;
  Container::iterator m;
  Container::iterator n;
  Container::iterator o;
  PointUCO posOrtogonal;
  PointUCO newAgentPosition;
  PointUCO oldAgentPosition;
  PointUCO newHelpPosition;
  VectorUCO vectorPasserAgent;
  VectorUCO vectorMovement;
  VectorUCO vectorPasserHelp;
  AngleDegUCO newAgentDirection;  
  AngleDegUCO angAgentOpp;
  AngleDegUCO angAgentBestOpp;
  FloatUCO distPasserAgent;
  FloatUCO distPasserOpp;
  FloatUCO distAgentOpp;
  FloatUCO distPasserBestOpp;
  FloatUCO distAgentBestOpp;
  FloatUCO distPasserOrtogonal;

  /*
   * Search for the teammate closest to the ball. ---
   */
  teammateList = playerView->getTeammates(NOW);
  i = teammateList.begin();
  j = teammateList.begin();
  while ( i != teammateList.end() ) 
    {
      if ( VectorUCO( (*i).getPosition(), ballView->getBallPosition(NOW) ).getLength() <
	   VectorUCO( (*j).getPosition(), ballView->getBallPosition(NOW) ).getLength() ) 
	{
	  j = i;
	}
      i++;
    }
  
  /* 
   * Check own position
   */
  oldAgentPosition = (agentView->getAgentPosition(NOW) );
  
  /* 
   * Check which opponent, l, is most likely to block pass.
   */
  opponentList = playerView->getOpponents(NOW);
  k = opponentList.begin();
  l = opponentList.begin();
  while ( k != opponentList.end() ) 
    {
      distPasserAgent = VectorUCO( (*j).getPosition(), oldAgentPosition ).getLength();
      distPasserOpp = VectorUCO( (*j).getPosition(), (*k).getPosition() ).getLength();
      distAgentOpp = VectorUCO( oldAgentPosition, (*k).getPosition() ).getLength();
      distPasserBestOpp = VectorUCO( (*j).getPosition(), (*l).getPosition() ).getLength();
      distAgentBestOpp = VectorUCO( oldAgentPosition, (*l).getPosition() ).getLength();
      
      /*
       * Check if opponent is closer to the ball than agent. In that case calculate angle.
       */
      if ( distPasserAgent > distPasserOpp ) 
	{	
	  angAgentOpp = (180/3.14) 
	    * acos(((distPasserAgent*distPasserAgent 
		     + distPasserOpp*distPasserOpp 
		     - distAgentOpp*distAgentOpp) 
		    / (2.0*distPasserAgent*distPasserOpp)).getValue());
	  angAgentBestOpp = (180/3.14) 
	    * acos(((distPasserAgent*distPasserAgent 
		     + distPasserBestOpp*distPasserBestOpp 
		     - distAgentBestOpp*distAgentBestOpp) 
		    / (2.0*distPasserAgent*distPasserBestOpp)).getValue()) ;
	  /* 
	   * Select from the opponents closer to the passer than the agent
	   * the one with the least angle.
	   */
	  if ( angAgentOpp.getDeg() < angAgentBestOpp.getDeg() ) 
	    { 
	      l = k;
	    }
	}
      k++;
    }
  
  /* 
   * Move to new posititon and turn to face passer.
   */ 
  vectorPasserAgent = VectorUCO( (*j).getPosition(), oldAgentPosition );
  distPasserBestOpp = VectorUCO( (*j).getPosition(), (*l).getPosition() ).getLength();
  distPasserOrtogonal = cos(angAgentBestOpp)*distPasserBestOpp; 
  posOrtogonal = (*j).getPosition() + 
    ((distPasserOrtogonal / distPasserAgent) * vectorPasserAgent );

  /* Get direction to helpPosition. */
  vectorMovement = VectorUCO( (*l).getPosition(), posOrtogonal );

  /* Set distance from oldAgentPosition to helpPosition to 5. */
  vectorMovement = vectorMovement.setLength(5);
  newHelpPosition = oldAgentPosition + vectorMovement;   
  vectorPasserHelp =  VectorUCO( (*j).getPosition(), newHelpPosition );

  /* Set newAgentPosition. It's a position away from opponent towards passer. */
  newAgentPosition = (*j).getPosition() + (0.8 * vectorPasserHelp); 

  /*
   * Check if newAgentPosition is outside any of the "kortsidor" In that case
   * move up from oldAgentPosition if passer is higher than opponent and down 
   * otherwise. 
   */
    if ( !(abs(newAgentPosition.getX()) < 53) ) 
      {
	if ( ((*j).getPosition()).getY() > ((*l).getPosition()).getY() ) 
	  { 
	    newAgentPosition = oldAgentPosition.addY(5); 
	  }
	else 
	  {
	    newAgentPosition = oldAgentPosition.addY(-5);
	  }
      }

  /*
   * Check if newAgentPosition is outside any of the "lngsidor" In that case
   * move forward from oldAgentPosition if passer is closer to opponent's goal 
   * than opponent and backward otherwise. 
   */    
    if ( !(abs(newAgentPosition.getY()) < 35) ) 
      {
	if ( ((*j).getPosition()).getX() > ((*l).getPosition()).getX() ) 
	  { 
	    newAgentPosition = oldAgentPosition.addX(5); 
	  }
	else 
	  {
	    newAgentPosition = oldAgentPosition.addX(-5);
	  }
      } 
    
    /*
     * Check if newAgentPosition still is outside the field. In that case
     * stay on oldAgentPosition.
     */
    if ( !( (abs (newAgentPosition.getX()) < 53) &&  
	    (abs (newAgentPosition.getY()) <35)) )
      {
	newAgentPosition = oldAgentPosition; 
      }
    
    /*
     * Check if newAgentPosition makes the agent offside. In that case
     * stay on oldAgentPosition.
     */
    opponentList = playerView->getOpponents(NOW);
    m = opponentList.begin();
    n = opponentList.begin(); 
    o = opponentList.begin();
    
    /* Get the opponent, o, who will cancel offside. */
    while ( m != opponentList.end() ) 
      {
	/* n is opponent closest to opponent's goal. */
	if ( ((*m).getPosition()).getX() < ((*n).getPosition()).getX() )
	  {
	    n = m;
	  } 
	
	/* n and o can not be same opponent. */    
	else if ( ((*n).getPosition()).getX() == ((*o).getPosition()).getX() )
	  {
	    o = m;
	  }
	
	/* o is opponent next closest to opponent's goal. */
	else if ( ((*m).getPosition()).getX() < ((*o).getPosition()).getX() )
	  {
	    o = m;
	  }
	m++; 
      }
    
    /* Check if newAgentPosition is offside. */
    if ( (newAgentPosition.getX() > 0)  &&  
	 (newAgentPosition.getX() > (((*o).getPosition()).getX())) )
      {
	newAgentPosition = oldAgentPosition;
      }
    
    /* 
     * Calculate direction to face passer. 
     */
    newAgentDirection = VectorUCO( newAgentPosition, (*j).getPosition() ).getAngle();
    
    /*
     * Move to newAgentPosition facing passer.
     */ 
    MoveToPositionFacingDirectionAvoidingObstacles *moveToPos = 
      new MoveToPositionFacingDirectionAvoidingObstacles();

    /* Check if it is possible to use skill "MoveToPosition...". */
    if ( moveToPos->applicable( newAgentPosition.getPoint(), 
				newAgentDirection.getDeg() ) )
      {
	return moveToPos->generatePlan( newAgentPosition.getPoint(), 
					newAgentDirection.getDeg() );
      }

    /* If not possible to use skill "MoveToPosition..." then stay in oldAgentPosition. */
    else
      {
	newAgentPosition = oldAgentPosition; 
	return moveToPos->generatePlan( newAgentPosition.getPoint(), 
					newAgentDirection.getDeg() );
      }
}


FuzzyBool MakePassable::persistent()
{
  return FuzzyBool(0.0);
}


FuzzyBool MakePassable::applicable()
{
  std::vector<PlayerObject> opponentList;
  std::vector<PlayerObject> teammateList;
  typedef std::vector<PlayerObject> Container;
  Container::iterator i;
  Container::iterator j;
  Container::iterator k;
  AngleDegUCO angAgentOpp;
  PointUCO oldAgentPosition; 
  FloatUCO distPasserAgent;
  FloatUCO distPasserOpp;
  FloatUCO distAgentOpp;

  /*
   * Check if before kick off.
   */
  if ( gameView->getPlayMode() == PM_BEFORE_KICK_OFF ) 
    {
      return FuzzyBool(0.0);
    }

  /*
   * Search for the teammate closest to the ball.
   */
  teammateList = playerView->getTeammates(NOW);
  if ( teammateList.empty() ) 
    { 
      return FuzzyBool(0.0);
    }
  i = teammateList.begin();
  j = teammateList.begin();
  while ( i != teammateList.end() ) 
    {
      if ( VectorUCO( (*i).getPosition(), ballView->getBallPosition(NOW) ).getLength() <
	   VectorUCO( (*j).getPosition(), ballView->getBallPosition(NOW) ).getLength() ) 
	{
	  j = i;
	}
      i++;
    }
  
  /* 
   * Check own position.
   */
  oldAgentPosition = (agentView->getAgentPosition(NOW) );
  
  /*
   * Check if any opponent is likely to block pass. 
   */
  opponentList = playerView->getOpponents(NOW);
  k = opponentList.begin();
  while ( k != opponentList.end() ) 
    {
      distPasserAgent = VectorUCO( (*j).getPosition(), oldAgentPosition ).getLength();
      distPasserOpp = VectorUCO( (*j).getPosition(), (*k).getPosition() ).getLength();
      distAgentOpp = VectorUCO( oldAgentPosition, (*k).getPosition() ).getLength();
      if ( distPasserAgent > distPasserOpp ) 
	{	
	  angAgentOpp = (180/3.14) 
	    * acos (((distPasserAgent*distPasserAgent 
		      + distPasserOpp*distPasserOpp 
		      - distAgentOpp*distAgentOpp) 
		     / (2.0*distPasserAgent*distPasserOpp)).getValue());
	  if ( angAgentOpp < 90 ) 
	    {
	      return FuzzyBool(1.0);
	    }
	}
      k++;
    }
  
  /* 
   * Return false if no opponent is threatening the pass. 
   */
  return FuzzyBool(0.0); 
  
}


RS_END_NAMESPACE





























