/*
 * $Id: TLine.cc,v 1.1 2001/09/18 20:41:49 andbo928 Exp $
 * 
 * HISTORY 
 *
 * $Log: TLine.cc,v $
 * Revision 1.1  2001/09/18 20:41:49  andbo928
 * Imported some sources from FuzzyFoo.
 *
 * Revision 1.4  2001/07/28 15:27:22  mikbr212
 * Added const
 *
 * Revision 1.3  2001/04/21 11:40:39  mikbr212
 * Made more methods inline
 *
 * Revision 1.2  2001/04/15 23:19:33  andbo928
 * It now ueses scalar products instead of component math.
 *
 * Revision 1.1  2001/04/13 16:17:56  mikbr212
 * converted from line
 *
 * Revision 1.1.1.1  2001/04/08 11:33:59  sigge
 * Imported sources
 *
 * Revision 1.7  2000/11/07 18:11:08  mikbr212
 * added methods
 *
 * Revision 1.6  2000/10/30 18:32:22  mikbr212
 * added comments
 *
 *
 */

/**
 * @file TLine.cc
 * Implementation file for the class TLine.
 *
 * $Revision: 1.1 $
 * $Author: andbo928 $
 * $Date: 2001/09/18 20:41:49 $
 */

#include "TLine.h"
#include <iostream.h>

RS_BEGIN_NAMESPACE

/**
 * Constructor.
 * Constructs line from first point to second point.
 */
TLine::TLine(const TVector& p1, const TVector& p2) 
  : p(p1),
    v(p2-p1)
{
}
/**
 * Constructor.
 * Constructs line from first point to tangent of circle.
 * @param p1 Starting point
 * @param c Center of circle
 * @param radius Radius of circle
 * @param clockwise True if a clockwise tangent is wanted (default is true)
 */

TLine::TLine(const TVector& p1, const TVector& c, const double& radius, bool clockwise)
  : p(p1)
{
  TVector pc = c-p;
  double alpha;
  if(pc.getLength() >= radius) { // starting point is outside circle
    alpha = ASin(radius / pc.getLength()).getDeg(); // in degrees
    pc.setLength(sqrt(pc.getLength()*pc.getLength() - radius*radius)); //radius / tan(alpha));
  } else { // starting point is in circle
    alpha = 90;
    pc.setLength(sqrt(radius*radius - pc.getLength()*pc.getLength()));
  }
  if(clockwise)
    pc.setAngle(pc.getAngle() - alpha);
  else
    pc.setAngle(pc.getAngle() + alpha);
  
  v = pc;
}
  
/**
 * Descructor.
 */
TLine::~TLine() {
}

/**
 * Checks if the lines are parallell.
 * @param l The line to compare to.
 * @returns True if the lines are parallell.
 */
bool TLine::isParallel(const TLine& l) const {

  double det = -v.getX() * l.v.getY() + l.v.getX() * v.getY();
  if(abs(det) < 0.00001)
    return true;
  return false;
}


/**
 * Calculates crosspoint of two lines.
 * @param l The other line.
 * @returns The crosspoint
 */
TVector TLine::getIntersection(const TLine& l) const {

  /* This sloves the equation system
   * p + v * t = l.p + l.v * s  =>
   * v * t - l.v * s = l.p - p
   *
   * Expressed as a matrix equation
   * 
   * | v.x -l.v.x | * |t| = |l.p.x - p.x|
   * | v.y -l.v.y |   |s| = |l.p.y - p.y|
   *
   * |a b|^-1 = ______1______ * |  d -b |
   * |c d|       (a*d - b*c)    | -c  a |
   */

  double a = v.getX();
  double b = -l.v.getX();
  double c = v.getY();
  double d = -l.v.getY();

  TVector dlt = l.p - p;

  double det = (a*d - b*c);
  if(abs(det) < 0.0001)
    return TVector(0,0);

  double t = ( d * dlt.getX() - b * dlt.getY()) / det;
  double s = (-c * dlt.getX() + a * dlt.getY()) / det;

//    cout << "p: " << p << ", l.p: " << l.p << endl;
//    cout << "v: " << v << ", l.v: " << l.v << endl;

  return p + t * v;
}

/**
 * Calculates crosspoint of two lines.
 * @param l The other line.
 * @param t Is set to the scaling-factor of vector where line l crosses this line.
 * ie if t is negative then the crosspoint is 'behind' the line, and if it is greater
 * than 1 it is in front of the line.
 * @param s Like t but for the other line.
 * @returns The crosspoint (starting point + t * vector = l.starting point + s * l.vector).
 */
TVector TLine::getIntersection(const TLine& l, double& t, double& s) const {
  double a = v.getX();
  double b = -l.v.getX();
  double c = v.getY();
  double d = -l.v.getY();

  TVector dlt = l.p - p;

  double det = (a*d - b*c);
  if(abs(det) < 0.0001)
    return TVector(0,0);

  t = ( d * dlt.getX() - b * dlt.getY()) / det;
  s = (-c * dlt.getX() + a * dlt.getY()) / det;

  return p + t * v;
}

/**
 * Extends (or crops) line to a circle.
 * If line (or it's extension) isn't through circle then nothing is done.
 * Note: The line will be mirrored of circle if is 'behind' the line.
 * @param c Center of circle
 * @param r Radius of circle
 */
bool TLine::extendToCircle(const TVector& c, const double& r) {
  double t,d,delta;
  d = getDistance(c,t).getLength();
  if(d > r)
    return false; // TLine (or it's extension) is not through circle

  delta = sqrt(r*r - d*d);

  if(delta + t * v.getLength() < 0) { // the direction is reversed
#ifdef LINE_SPAM    
    cout << "********* TLine is mirrored! *********\n";
#endif
    v = -v; // mirror
    v.setLength(-(delta + t * v.getLength()));
  } else {
    v.setLength(delta + t * v.getLength());
  }
  
  return true;
}

RS_END_NAMESPACE

/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler or debugger does not understand the keyword inline then
 * include the inline definitions here, otherwise include them in the
 * declaration file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifndef RS_USE_INLINE
#  include "TLine.icc"
#endif
